@@ -1,6 +1,7 @@
 /*
  * LibGCab
  * Copyright (c) 2012, Marc-André Lureau <marcandre.lureau@gmail.com>
+ * Copyright (c) 2017, Richard Hughes <richard@hughsie.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -44,9 +45,15 @@
  * gcab_folder_add_file().
  */
 
-struct _GCabFolderClass
+struct _GCabFolder
 {
-    GObjectClass parent_class;
+    GObject parent_instance;
+
+    GSList *files;
+    GHashTable *hash;
+    gint comptype;
+    GByteArray *reserved;
+    cfolder_t *cfolder;
 };
 
 enum {
@@ -71,12 +78,11 @@ gcab_folder_finalize (GObject *object)
 {
     GCabFolder *self = GCAB_FOLDER (object);
 
+    cfolder_free (self->cfolder);
     g_slist_free_full (self->files, g_object_unref);
     g_hash_table_unref (self->hash);
     if (self->reserved)
         g_byte_array_unref (self->reserved);
-    if (self->stream)
-        g_object_unref (self->stream);
 
     G_OBJECT_CLASS (gcab_folder_parent_class)->finalize (object);
 }
@@ -158,14 +164,31 @@ G_GNUC_INTERNAL gsize
 gcab_folder_get_ndatablocks (GCabFolder *self)
 {
     gsize total_size = 0;
-    GSList *l;
 
-    for (l = self->files; l != NULL; l = l->next)
-        total_size += GCAB_FILE (l->data)->cfile.usize;
+    for (GSList *l = self->files; l != NULL; l = l->next) {
+        GCabFile *file = GCAB_FILE (l->data);
+        total_size += gcab_file_get_usize (file);
+    }
 
     return total_size / DATABLOCKSIZE + 1 ;
 }
 
+/**
+ * gcab_folder_get_comptype:
+ * @cabfolder: a #GCabFolder
+ *
+ * Returns the compression used in this folder.
+ *
+ * Returns: a #GCabCompression, e.g. %GCAB_COMPRESSION_MSZIP
+ *
+ * Since: 1.0
+ **/
+gint
+gcab_folder_get_comptype (GCabFolder *self)
+{
+    return self->comptype;
+}
+
 static gboolean
 add_file (GCabFolder *self, GCabFile *file, GError **error)
 {
@@ -184,6 +207,14 @@ add_file (GCabFolder *self, GCabFile *file, GError **error)
 
 #define FILE_ATTRS "standard::*,time::modified"
 
+static gint
+_sort_cfiles (gconstpointer a, gconstpointer b)
+{
+    GCabFile *file_a = GCAB_FILE (a);
+    GCabFile *file_b = GCAB_FILE (b);
+    return g_strcmp0 (gcab_file_get_name (file_a), gcab_file_get_name (file_b));
+}
+
 static gboolean
 add_file_info (GCabFolder *self, GCabFile *file, GFileInfo *info,
                const gchar *name, gboolean recurse, GError **error)
@@ -194,33 +225,28 @@ add_file_info (GCabFolder *self, GCabFile *file, GFileInfo *info,
         if (!recurse)
             return TRUE;
 
-        GFileEnumerator *dir = g_file_enumerate_children (file->file, FILE_ATTRS, 0, NULL, error);
-        if (*error) {
+        g_autoptr(GFileEnumerator) dir = g_file_enumerate_children (gcab_file_get_gfile (file), FILE_ATTRS, 0, NULL, error);
+        if (dir == NULL) {
             g_warning ("Couldn't enumerate directory %s: %s", name, (*error)->message);
             g_clear_error (error);
             return TRUE;
         }
 
         while ((info = g_file_enumerator_next_file (dir, NULL, error)) != NULL) {
-            GFile *child = g_file_get_child (file->file, g_file_info_get_name (info));
-            gchar *child_name = g_build_path ("\\", name, g_file_info_get_name (info), NULL);
-            GCabFile *child_file = gcab_file_new_with_file (child_name, child);
-
-            add_file_info (self, child_file, info, child_name, recurse, error);
-            if (*error) {
-                g_warning ("Couldn't add file %s: %s",
-                           child_name, (*error)->message);
-                g_clear_error (error);
+            g_autoptr(GFile) child = g_file_get_child (gcab_file_get_gfile (file), g_file_info_get_name (info));
+            g_autofree gchar *child_name = g_build_path ("\\", name, g_file_info_get_name (info), NULL);
+            g_autoptr(GCabFile) child_file = gcab_file_new_with_file (child_name, child);
+            if (!add_file_info (self, child_file, info, child_name, recurse, error)) {
+                g_object_unref (info);
+                return FALSE;
             }
 
-            g_object_unref (child_file);
-            g_free (child_name);
-            g_object_unref (child);
+            /* sort the files to avoid depending on filesystem order */
+            self->files = g_slist_sort (self->files, _sort_cfiles);
+
             g_object_unref (info);
         }
 
-        g_object_unref (dir);
-
     } else if (file_type == G_FILE_TYPE_REGULAR) {
         gcab_file_update_info (file, info);
         if (!add_file (self, file, error))
@@ -263,13 +289,12 @@ gcab_folder_add_file (GCabFolder *self, GCabFile *file,
     if (gfile) {
         g_return_val_if_fail (G_IS_FILE (gfile), FALSE);
 
-        GFileInfo *info = g_file_query_info (gfile, FILE_ATTRS, 0, NULL, error);
+        g_autoptr(GFileInfo) info = g_file_query_info (gfile, FILE_ATTRS, 0, NULL, error);
         if (info == NULL)
             return FALSE;
 
         success = add_file_info (self, file, info,
                                  gcab_file_get_name (file), recurse, error);
-        g_object_unref (info);
     } else {
         success = add_file (self, file, error);
     }
@@ -313,13 +338,14 @@ gcab_folder_new (gint comptype)
 }
 
 G_GNUC_INTERNAL GCabFolder *
-gcab_folder_new_with_cfolder (const cfolder_t *folder, GInputStream *stream)
+gcab_folder_new_steal_cfolder (cfolder_t **cfolder)
 {
+    g_return_val_if_fail (cfolder != NULL, NULL);
+
     GCabFolder *self = g_object_new (GCAB_TYPE_FOLDER,
-                                     "comptype", folder->typecomp,
+                                     "comptype", (*cfolder)->typecomp,
                                      NULL);
-    self->stream = g_object_ref (stream);
-    self->cfolder = *folder;
+    self->cfolder = g_steal_pointer (cfolder);
 
     return self;
 }
@@ -340,18 +366,54 @@ gcab_folder_get_files (GCabFolder *self)
     return g_slist_reverse (g_slist_copy (self->files));
 }
 
+/**
+ * gcab_folder_get_file_by_name:
+ * @cabfolder: a #GCabFolder
+ * @name: a file name
+ *
+ * Gets a specific #GCabFile files contained in the @cabfolder.
+ *
+ * Returns: (transfer none): A #GCabFile, or %NULL if not found
+ **/
+GCabFile *
+gcab_folder_get_file_by_name (GCabFolder *self, const gchar *name)
+{
+    GCabFile *cabfile;
+
+    g_return_val_if_fail (GCAB_IS_FOLDER (self), NULL);
+    g_return_val_if_fail (name != NULL, NULL);
+
+    /* try the hash first */
+    cabfile = g_hash_table_lookup (self->hash, name);
+    if (cabfile != NULL)
+        return cabfile;
+
+    /* if the extract name is different, look for that too */
+    for (GSList *l = self->files; l != NULL; l = l->next) {
+        cabfile = GCAB_FILE (l->data);
+        if (gcab_file_get_name (cabfile) != gcab_file_get_extract_name (cabfile)) {
+            if (g_strcmp0 (gcab_file_get_extract_name (cabfile), name) == 0)
+                return cabfile;
+        }
+    }
+
+    /* nothing found */
+    return NULL;
+}
+
 static gint
 sort_by_offset (GCabFile *a, GCabFile *b)
 {
     g_return_val_if_fail (a != NULL, 0);
     g_return_val_if_fail (b != NULL, 0);
 
-    return (gint64)a->cfile.uoffset - (gint64)b->cfile.uoffset;
+    return (gint64) gcab_file_get_uoffset (a) - (gint64) gcab_file_get_uoffset (b);
 }
 
 G_GNUC_INTERNAL gboolean
 gcab_folder_extract (GCabFolder *self,
-                     GFile *path,
+                     GDataInputStream *data,
+                     GFile *path_extract,
                      guint8 res_data,
                      GCabFileCallback file_callback,
                      GFileProgressCallback progress_callback,
@@ -359,114 +421,75 @@ gcab_folder_extract (GCabFolder *self,
                      GCancellable *cancellable,
                      GError **error)
 {
-    GError *my_error = NULL;
-    gboolean success = FALSE;
-    GDataInputStream *data = NULL;
-    GFileOutputStream *out = NULL;
-    GSList *f, *files = NULL;
-    cdata_t cdata = { 0, };
+    GSList *f = NULL;
+    g_autoptr(GSList) files = NULL;
+    g_autoptr(cdata_t) cdata = g_new0 (cdata_t, 1);
     guint32 nubytes = 0;
 
-    data = g_data_input_stream_new (self->stream);
-    g_data_input_stream_set_byte_order (data, G_DATA_STREAM_BYTE_ORDER_LITTLE_ENDIAN);
-    g_filter_input_stream_set_close_base_stream (G_FILTER_INPUT_STREAM (data), FALSE);
+    /* never loaded from a stream */
+    g_assert (self->cfolder != NULL);
 
-    if (!g_seekable_seek (G_SEEKABLE (data), self->cfolder.offsetdata, G_SEEK_SET, cancellable, error))
-        goto end;
+    if (!g_seekable_seek (G_SEEKABLE (data), self->cfolder->offsetdata, G_SEEK_SET, cancellable, error))
+        return FALSE;
 
     files = g_slist_sort (g_slist_copy (self->files), (GCompareFunc)sort_by_offset);
 
+    /* this is allocated for every block, but currently unused */
+    cdata->reserved = g_malloc (res_data);
+
     for (f = files; f != NULL; f = f->next) {
         GCabFile *file = f->data;
 
         if (file_callback && !file_callback (file, callback_data))
             continue;
 
-        gchar *fname = g_strdup (gcab_file_get_extract_name (file));
-        int i = 0, len = strlen (fname);
-        for (i = 0; i < len; i++)
-            if (fname[i] == '\\')
-                fname[i] = '/';
+        g_autoptr(GOutputStream) out = NULL;
+        out = gcab_file_get_output_stream (file, path_extract, cancellable, error);
+        if (out == NULL)
+            return FALSE;
 
-        GFile *gfile = g_file_resolve_relative_path (path, fname);
-        g_free (fname);
-
-        if (!g_file_has_prefix (gfile, path)) {
-            // "Rebase" the file in the given path, to ensure we never escape it
-            char *rawpath = g_file_get_path (gfile);
-            if (rawpath != NULL) {
-                char *newpath = rawpath;
-                while (*newpath != 0 && *newpath == G_DIR_SEPARATOR) {
-                    newpath++;
-                }
-                GFile *newgfile = g_file_resolve_relative_path (path, newpath);
-                g_free (rawpath);
-                g_object_unref (gfile);
-                gfile = newgfile;
-            }
-        }
-
-        GFile *parent = g_file_get_parent (gfile);
-
-        if (!g_file_make_directory_with_parents (parent, cancellable, &my_error)) {
-            if (g_error_matches (my_error, G_IO_ERROR, G_IO_ERROR_EXISTS))
-                g_clear_error (&my_error);
-            else {
-                g_object_unref (gfile);
-                g_object_unref (parent);
-                g_propagate_error (error, my_error);
-                goto end;
-            }
-        }
-        g_object_unref (parent);
-
-        g_clear_object (&out);
-        out = g_file_replace (gfile, NULL, FALSE, G_FILE_CREATE_REPLACE_DESTINATION, cancellable, error);
-        g_object_unref (gfile);
-        if (!out)
-            goto end;
-
-        guint32 usize = file->cfile.usize;
-        guint32 uoffset = file->cfile.uoffset;
+        guint32 usize = gcab_file_get_usize (file);
+        guint32 uoffset = gcab_file_get_uoffset (file);
 
         /* let's rewind if need be */
         if (uoffset < nubytes) {
-            if (!g_seekable_seek (G_SEEKABLE (data), self->cfolder.offsetdata,
+            if (!g_seekable_seek (G_SEEKABLE (data), self->cfolder->offsetdata,
                                   G_SEEK_SET, cancellable, error))
-                goto end;
-            bzero(&cdata, sizeof(cdata));
+                return FALSE;
             nubytes = 0;
         }
 
         while (usize > 0) {
-            if ((nubytes + cdata.nubytes) <= uoffset) {
-                nubytes += cdata.nubytes;
-                if (!cdata_read (&cdata, res_data, self->comptype,
+            if ((nubytes + cdata->nubytes) <= uoffset) {
+                nubytes += cdata->nubytes;
+                if (!cdata_read (cdata, res_data, self->comptype,
                                  data, cancellable, error))
-                    goto end;
+                    return FALSE;
                 continue;
             } else {
-                gsize offset = file->cfile.uoffset > nubytes ?
-                    file->cfile.uoffset - nubytes : 0;
-                const void *p = &cdata.out[offset];
-                gsize count = MIN (usize, cdata.nubytes - offset);
+                gsize offset = gcab_file_get_uoffset (file) > nubytes ?
+                    gcab_file_get_uoffset (file) - nubytes : 0;
+                const void *p = &cdata->out[offset];
+                gsize count = MIN (usize, cdata->nubytes - offset);
                 if (!g_output_stream_write_all (G_OUTPUT_STREAM (out), p, count,
                                                 NULL, cancellable, error))
-                    goto end;
+                    return FALSE;
                 usize -= count;
                 uoffset += count;
             }
         }
+
+        if (!g_output_stream_close (out, cancellable, error))
+            return FALSE;
+
+        /* no backing GFile */
+        if (path_extract == NULL) {
+            g_autoptr(GBytes) blob = NULL;
+            blob = g_memory_output_stream_steal_as_bytes (G_MEMORY_OUTPUT_STREAM (out));
+            gcab_file_set_bytes (file, blob);
+        }
+
     }
 
-    success = TRUE;
-
-end:
-    g_slist_free (files);
-
-    g_clear_object (&data);
-    g_clear_object (&out);
-    cdata_finish (&cdata, NULL);
-
-    return success;
+    return TRUE;
 }
